/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package haproxy

import (
	"bytes"
	"io"
	"io/ioutil"
	"net/http"
	"net/url"
	"strconv"
	"time"

	"github.com/go-openapi/strfmt"
	"github.com/haproxytech/models"
)

var fmts = strfmt.NewFormats()

const (
	dataplaneUser    = "dataplaneapi"
	txPath           = "/v1/services/haproxy/transactions"
	sitesPath        = "/v1/services/haproxy/sites"
	reloadsPath      = "/v1/services/haproxy/reloads"
	versionHdr       = "Configuration-Version"
	reloadIdHdr      = "Reload-ID"
	varnishSock      = "unix@/varnish.sock"
	frontend         = "offloader"
	backend          = "varnish"
	server           = backend
	frontendSitePath = sitesPath + "/" + frontend
)

var port = int64(443)

type ReloadStatus uint8

const (
	Unknown ReloadStatus = iota
	Failed
	InProgress
	Succeeded
)

func status(reloadStr string) ReloadStatus {
	switch reloadStr {
	case "failed":
		return Failed
	case "in_progress":
		return InProgress
	case "succeeded":
		return Succeeded
	default:
		return Unknown
	}
}

func (status ReloadStatus) String() string {
	switch status {
	case Failed:
		return "failed"
	case InProgress:
		return "in_progress"
	case Succeeded:
		return "succeeded"
	default:
		return "UNKNOWN"
	}
}

type ReloadState struct {
	ID        string
	Response  string
	Timestamp time.Time
	Status    ReloadStatus
}

type DataplaneError struct {
	Err     *models.Error
	Status  int
	Version int
}

func (err *DataplaneError) Error() string {
	return *err.Err.Message
}

type DataplaneClient struct {
	baseURL  *url.URL
	user     string
	password string
	client   *http.Client
}

func NewDataplaneClient(host, pass string) *DataplaneClient {
	return &DataplaneClient{
		baseURL: &url.URL{
			Scheme: "http",
			Host:   host,
		},
		user:     dataplaneUser,
		password: pass,
		client:   http.DefaultClient,
	}
}

func (client *DataplaneClient) getHost() string {
	return client.baseURL.Host
}

func getSite(certName string) ([]byte, error) {
	site := &models.Site{
		Name: frontend,
		Service: &models.SiteService{
			Listeners: []*models.Bind{
				&models.Bind{
					Name:           frontend,
					Port:           &port,
					Ssl:            true,
					SslCertificate: certName,
				},
			},
		},
		Farms: []*models.SiteFarm{
			&models.SiteFarm{
				Name:  backend,
				UseAs: "default",
				Servers: []*models.Server{
					&models.Server{
						Name:        server,
						Address:     varnishSock,
						Check:       "enabled",
						SendProxyV2: "enabled",
					},
				},
			},
		},
	}
	return site.MarshalBinary()
}

func drainAndClose(body io.ReadCloser) {
	io.Copy(ioutil.Discard, body)
	body.Close()
}

func (client *DataplaneClient) getReq(
	path, method string,
	body io.Reader,
	withbody bool,
) (req *http.Request, err error) {
	relPath := &url.URL{Path: path}
	url := client.baseURL.ResolveReference(relPath)
	if withbody {
		req, err = http.NewRequest(method, url.String(), body)
	} else {
		req, err = http.NewRequest(method, url.String(), nil)
	}
	if err != nil {
		return req, err
	}
	req.SetBasicAuth(client.user, client.password)
	req.Header.Set("Accept", "application/json")
	return req, nil
}

func setVersion(req *http.Request, version int64) {
	query := req.URL.Query()
	query.Add("version", strconv.FormatInt(version, 10))
	req.URL.RawQuery = query.Encode()
}

func setTx(req *http.Request, tx *models.Transaction) {
	query := req.URL.Query()
	query.Add("transaction_id", tx.ID)
	req.URL.RawQuery = query.Encode()
}

func getBody(resp *http.Response) ([]byte, error) {
	defer drainAndClose(resp.Body)
	// XXX read ContentLength bytes
	return ioutil.ReadAll(resp.Body)
}

func getTx(body []byte) (tx *models.Transaction, err error) {
	tx = &models.Transaction{}
	if err = tx.UnmarshalBinary(body); err != nil {
		err = tx.Validate(fmts)
	}
	return tx, err
}

func getError(resp *http.Response, body []byte) error {
	dplaneErr := &DataplaneError{
		Err:    &models.Error{},
		Status: resp.StatusCode,
	}
	if dplaneVer, err := strconv.Atoi(
		resp.Header.Get(versionHdr)); err != nil {
		dplaneErr.Version = dplaneVer
	}
	dplaneErr.Err.UnmarshalJSON(body)
	return dplaneErr
}

func (client *DataplaneClient) StartTx(
	version int64) (tx *models.Transaction, err error) {

	// path := &url.URL{Path: "/v1/services/haproxy/transactions"}
	// url := client.baseURL.ResolveReference(path)
	// req, err := http.NewRequest("POST", url.String(), nil)
	// if err != nil {
	// 	return
	// }
	// req.SetBasicAuth(client.user, client.password)
	// query := req.URL.Query()
	// query.Add("version", strconv.FormatInt(version, 10))
	// req.URL.RawQuery = query.Encode()
	// req.Header.Set("Accept", "application/json")

	req, err := client.getReq(txPath, http.MethodPost, nil, false)
	if err != nil {
		return
	}
	setVersion(req, version)
	resp, err := client.client.Do(req)
	if err != nil {
		return
	}
	// defer drainAndClose(resp.Body)
	// // XXX read ContentLength bytes
	// body, err := ioutil.ReadAll(resp.Body)
	// if err != nil {
	// 	return
	// }
	body, err := getBody(resp)
	if err != nil {
		return
	}

	if resp.StatusCode == http.StatusCreated {
		// tx = &models.Transaction{}
		// if err = tx.UnmarshalBinary(body); err != nil {
		// 	err = tx.Validate(fmts)
		// }
		// return
		return getTx(body)
	}

	// dplaneErr := &DataplaneError{
	// 	Err:    &models.Error{},
	// 	Status: resp.StatusCode,
	// }
	// if dplaneVer, err := strconv.Atoi(
	// 	resp.Header.Get("Configuration-Version")); err != nil {
	// 	dplaneErr.Version = dplaneVer
	// }
	// dplaneErr.Err.UnmarshalJSON(body)
	// return nil, dplaneErr
	return nil, getError(resp, body)
}

func (client *DataplaneClient) FinishTx(
	tx *models.Transaction) (ReloadState, error) {
	// path := &url.URL{Path: "/v1/services/haproxy/transactions/" + tx.ID}
	// url := client.baseURL.ResolveReference(path)
	// req, err := http.NewRequest("PUT", url.String(), nil)
	// if err != nil {
	// 	return false, err
	// }
	// req.SetBasicAuth(client.user, client.password)
	// req.Header.Set("Accept", "application/json")

	state := ReloadState{Status: Unknown}
	req, err := client.getReq(txPath+"/"+tx.ID, http.MethodPut, nil, false)
	if err != nil {
		return state, err
	}
	setVersion(req, tx.Version)
	resp, err := client.client.Do(req)
	if err != nil {
		return state, err
	}
	// defer drainAndClose(resp.Body)
	// // XXX read ContentLength bytes
	// body, err := ioutil.ReadAll(resp.Body)
	// if err != nil {
	// 	return false, err
	// }
	body, err := getBody(resp)
	if err != nil {
		return state, err
	}

	switch resp.StatusCode {
	case http.StatusOK, http.StatusAccepted:
		if tx, err = getTx(body); err != nil {
			return state, err
		}
		if resp.StatusCode == http.StatusOK {
			state.Status = Succeeded
			state.Timestamp = time.Now()
			return state, nil
		}
		state.ID = resp.Header.Get(reloadIdHdr)
		state.Status = status(tx.Status)
		return state, nil
	default:
		return state, getError(resp, body)
	}

	// if resp.StatusCode == http.StatusOK || resp.StatusCode == http.StatusAccepted {
	// 	if tx, err = body.getTx(); err != nil {
	// 		return false, err
	// 	}
	// 	return resp.StatusCode == http.StatusOK, nil
	// }

	// dplaneErr := &DataplaneError{
	// 	Err:    &models.Error{},
	// 	Status: resp.StatusCode,
	// }
	// if dplaneVer, err := strconv.Atoi(
	// 	resp.Header.Get("Configuration-Version")); err != nil {
	// 	dplaneErr.Version = dplaneVer
	// }
	// dplaneErr.Err.UnmarshalJSON(body)
	// return false, dplaneErr
	// return false, resp.getError(body)
}

func (client *DataplaneClient) configTLS(
	tx *models.Transaction, spec Spec, path, method string) error {
	// port := int64(443)
	// site := &models.Site{
	// 	Name: "offloader",
	// 	Service: &models.SiteService{
	// 		Listeners: []*models.Bind{
	// 			&models.Bind{
	// 				Name:           "offloader",
	// 				Port:           &port,
	// 				Ssl:            true,
	// 				SslCertificate: "ssl-cert-snakeoil.pem",
	// 			},
	// 		},
	// 	},
	// 	Farms: []*models.SiteFarm{
	// 		&models.SiteFarm{
	// 			Name:  "varnish",
	// 			UseAs: "default",
	// 			Servers: []*models.Server{
	// 				&models.Server{
	// 					Name:        "varnish",
	// 					Address:     varnishSock,
	// 					Check:       "enabled",
	// 					SendProxyV2: "enabled",
	// 				},
	// 			},
	// 		},
	// 	},
	// }
	// siteBytes, err := site.MarshalBinary()
	// if err != nil {
	// 	return err
	// }
	var rdr *bytes.Reader
	if method != http.MethodDelete {
		site, err := getSite(spec.CertName())
		if err != nil {
			return err
		}
		rdr = bytes.NewReader(site)
	}

	// path := &url.URL{Path: "/v1/services/haproxy/sites"}
	// url := client.baseURL.ResolveReference(path)
	// req, err := http.NewRequest("POST", url.String(),
	// 	bytes.NewReader(siteBytes))
	// if err != nil {
	// 	return err
	// }
	// query := req.URL.Query()
	// query.Add("transaction_id", tx.ID)
	// req.URL.RawQuery = query.Encode()
	// req.SetBasicAuth(client.user, client.password)
	// req.Header.Set("Accept", "application/json")
	req, err := client.getReq(path, method, rdr,
		method != http.MethodDelete)
	if err != nil {
		return err
	}
	setTx(req, tx)
	if method != http.MethodDelete {
		req.Header.Set("Content-Type", "application/json")
	}

	resp, err := client.client.Do(req)
	if err != nil {
		return err
	}
	// defer drainAndClose(resp.Body)
	// // XXX read ContentLength bytes
	// body, err := ioutil.ReadAll(resp.Body)
	// if err != nil {
	// 	return err
	// }
	body, err := getBody(resp)
	if err != nil {
		return err
	}

	switch resp.StatusCode {
	case http.StatusAccepted:
		// XXX return the Reload-ID?
		if method != http.MethodDelete {
			// Validate the response body
			site := &models.Site{}
			if err = site.UnmarshalBinary(body); err != nil {
				return err
			}
			if err = site.Validate(fmts); err != nil {
				return err
			}
		}
		return nil
	case http.StatusOK, http.StatusCreated, http.StatusNoContent:
		panic("Got " + http.StatusText(resp.StatusCode) +
			" but did not set force_reload")
	default:
		return getError(resp, body)
	}

	// if resp.StatusCode == http.StatusCreated {
	// 	// XXX ????
	// 	panic("Got Created but did not set force_reload")
	// }
	// if resp.StatusCode != http.StatusAccepted {
	// 	dplaneErr := &DataplaneError{
	// 		Err:    &models.Error{},
	// 		Status: resp.StatusCode,
	// 	}
	// 	if dplaneVer, err := strconv.Atoi(
	// 		resp.Header.Get("Configuration-Version")); err != nil {
	// 		dplaneErr.Version = dplaneVer
	// 	}
	// 	dplaneErr.Err.UnmarshalJSON(body)
	// 	return dplaneErr
	// }

	// site = &models.Site{}
	// if err = site.UnmarshalBinary(body); err != nil {
	// 	return err
	// }
	// if err = site.Validate(fmts); err != nil {
	// 	return err
	// }
	// return nil
}

func (client *DataplaneClient) AddTLS(
	tx *models.Transaction, spec Spec) error {

	return client.configTLS(tx, spec, sitesPath, http.MethodPost)
}

func (client *DataplaneClient) UpdateTLS(
	tx *models.Transaction, spec Spec) error {

	return client.configTLS(tx, spec, frontendSitePath, http.MethodPut)
}

func (client *DataplaneClient) DeleteTLS(tx *models.Transaction) error {

	return client.configTLS(tx, Spec{}, frontendSitePath, http.MethodDelete)
}

func (client *DataplaneClient) DeleteTx(tx *models.Transaction) error {
	return nil
}

func (client *DataplaneClient) Reloaded(id string) (bool, ReloadState, error) {
	state := ReloadState{ID: id}
	req, err := client.getReq(reloadsPath+"/"+id, http.MethodGet, nil,
		false)
	if err != nil {
		return false, state, err
	}
	resp, err := client.client.Do(req)
	if err != nil {
		return false, state, err
	}
	body, err := getBody(resp)
	if err != nil {
		return false, state, err
	}

	switch resp.StatusCode {
	case http.StatusOK:
		reload := &models.Reload{}
		if err = reload.UnmarshalBinary(body); err != nil {
			return false, state, err
		}
		if err = reload.Validate(fmts); err != nil {
			return false, state, err
		}
		state.ID = reload.ID
		state.Response = reload.Response
		state.Timestamp = time.Unix(reload.ReloadTimestamp, 0)
		state.Status = status(reload.Status)
		return true, state, nil
	default:
		dplaneErr := getError(resp, body)
		if resp.StatusCode == http.StatusNotFound {
			return false, state, nil
		}
		return false, state, dplaneErr
	}
}
